@@ -26,6 +26,22 @@
 #define _(s) gettext (s)
 #define N_(s) s
 
+/* The width in bits of the integer type or expression T. */
+#define TYPE_WIDTH(t) (sizeof (t) * CHAR_BIT)
+
+#define SIGNED_TYPE_MAXIMUM(t) \
+  ((t) ((((t) 1 << (TYPE_WIDTH (t) - 2)) - 1) * 2 + 1))
+
+/* Maximum value for off_t */
+#define OFF_T_MAX SIGNED_TYPE_MAXIMUM (off_t)
+
+/* Return true if A can be added to B without integer overflow */
+static inline off_t
+off_t_sum_ok (off_t a, off_t b)
+{
+  return OFF_T_MAX - a >= b;
+}
+
 /* The type definitions are next.  */
 
 /* The available file space is stored in an "avail" table.  The one with
@@ -51,6 +67,17 @@ typedef struct
   avail_elem av_table[1]; /* The table.  Make it look like an array.  */
 } avail_block;
 
+/* Return true if avail_block is valid */
+static int inline
+gdbm_avail_block_valid_p (avail_block const *av)
+{
+  return (av->size > 1 && av->count >= 0 && av->count <= av->size);
+}
+
+/* Return true if both AV and the size of AV->av_table are valid */
+extern int gdbm_avail_table_valid_p (GDBM_FILE dbf,
+				     avail_elem const *av, int count);
+
 /* The dbm file header keeps track of the current location of the hash
    directory and the free space in the file.  */
 
@@ -86,6 +113,7 @@ typedef struct
   int   data_size;        /* Size of associated data in the file. */
 } bucket_element;
 
+extern int gdbm_bucket_element_valid_p (GDBM_FILE dbf, int elem_loc);
 
 /* A bucket is a small hash table.  This one consists of a number of
    bucket elements plus some bookkeeping fields.  The number of elements
@@ -121,11 +149,12 @@ typedef struct
 
 typedef struct
 {
-  int   hash_val;
-  int   data_size;
-  int   key_size;
-  char *dptr;
-  int   elem_loc;
+  int     hash_val;
+  int     data_size;
+  int     key_size;
+  char    *dptr;
+  size_t  dsize;
+  int     elem_loc;
 } data_cache_elem;
 
 typedef struct
@@ -198,7 +227,7 @@ struct gdbm_file_info
   /* The bucket cache. */
   cache_elem *bucket_cache;
   size_t cache_size;
-  int last_read;
+  size_t last_read;
 
   /* Points to the current hash bucket in the cache. */
   hash_bucket *bucket;
@@ -297,23 +326,9 @@ struct gdbm_file_info
       gdbm_set_errno(dbf, ec, fatal);                                   \
     }                                                                   \
   while (0)
-
-typedef int (*gdbm_debug_hook) (char const *, int, char const *, void *);
-extern void _gdbm_debug_hook_install (char const *, gdbm_debug_hook, void *);
-extern void _gdbm_debug_hook_remove (char const *);
-extern int _gdbm_debug_hook_check (char const *, int, char const *);
-extern int _gdbm_debug_hook_val (char const *);
-# define GDBM_DEBUG_HOOK(id) _gdbm_debug_hook_check(__FILE__,__LINE__,id)
-# define GDBM_DEBUG_OVERRIDE(id, stmt)				\
-  (GDBM_DEBUG_HOOK(id) ? _gdbm_debug_hook_val(id) : (stmt))
-# define GDBM_DEBUG_ALLOC(id, stmt)		\
-  (GDBM_DEBUG_HOOK(id) ? NULL : (stmt))
 #else
 # define GDBM_DEBUG(flags, fmt, ...)
 # define GDBM_DEBUG_DATUM(flags, dat, fmt, ...)
-# define GDBM_DEBUG_HOOK(id) 0
-# define GDBM_DEBUG_OVERRIDE(id, stmt) (stmt)
-# define GDBM_DEBUG_ALLOC(id, stmt) (stmt)
 # define GDBM_SET_ERRNO2(dbf, ec, fatal, m) gdbm_set_errno (dbf, ec, fatal)
 #endif
 
