@@ -103,6 +103,8 @@ _gdbm_dump_ascii (GDBM_FILE dbf, FILE *fp)
 	      break;
 	    }
  	}
+      else
+	break;
       nextkey = gdbm_nextkey (dbf, key);
       free (key.dptr);
       free (data.dptr);
@@ -110,8 +112,9 @@ _gdbm_dump_ascii (GDBM_FILE dbf, FILE *fp)
       count++;
     }
 
-  if (rc == 0)
+  if (rc == 0 && (rc = gdbm_last_errno (dbf)) == 0)
     {
+      
       /* FIXME: Something like that won't hurt, although load does not
 	 use it currently. */
       fprintf (fp, "#:count=%lu\n", (unsigned long) count);
