@@ -1,7 +1,7 @@
 /* gdbmexp.c - Export a GDBM database. */
 
 /* This file is part of GDBM, the GNU data base manager.
-   Copyright (C) 2007, 2011, 2013, 2016-2017 Free Software Foundation,
+   Copyright (C) 2007, 2011, 2013, 2016-2018 Free Software Foundation,
    Inc.
 
    GDBM is free software; you can redistribute it and/or modify
@@ -21,8 +21,13 @@
 # include "autoconf.h"
 # include <arpa/inet.h>
 
+#ifdef GDBM_EXPORT_18
+# define GDBM_SET_ERRNO(dbf, ec, fatal) gdbm_errno = ec
+# define GDBM_ASSERT_CONSISTENCY(dbf, val)
+#else
 # include "gdbmdefs.h"
 # include "gdbm.h"
+#endif
 
 int
 gdbm_export_to_file (GDBM_FILE dbf, FILE *fp)
