@@ -21,6 +21,7 @@
 #include "autoconf.h"
 
 #include "gdbmdefs.h"
+#include <stddef.h>
 
 /* Determine our native magic number and bail if we can't. */
 #if SIZEOF_OFF_T == 4
@@ -32,7 +33,7 @@
 #endif
 
 static void
-compute_directory_size (GDBM_FILE dbf, blksize_t block_size,
+compute_directory_size (blksize_t block_size,
 			int *ret_dir_size, int *ret_dir_bits)
 {
   /* Create the initial hash table directory.  */
@@ -49,6 +50,123 @@ compute_directory_size (GDBM_FILE dbf, blksize_t block_size,
   *ret_dir_bits = dir_bits;
 }
 
+static inline int
+bucket_element_count (size_t bucket_size)
+{
+  return (bucket_size - sizeof (hash_bucket)) / sizeof (bucket_element) + 1;
+}
+
+int
+gdbm_avail_table_valid_p (GDBM_FILE dbf, avail_elem const *av, int count)
+{
+  off_t prev = 0;
+  int i;
+    
+  prev = 0;
+  for (i = 0; i < count; i++, av++)
+    {
+      if (!(av->av_size >= prev
+	    && av->av_adr >= dbf->header->bucket_size
+	    && av->av_adr + av->av_size <= dbf->header->next_block))
+	return 0;
+      prev = av->av_size;
+    }
+  return 1;
+}
+
+int
+gdbm_avail_block_validate (GDBM_FILE dbf, avail_block *avblk)
+{
+  if (!(gdbm_avail_block_valid_p (avblk)
+	&& gdbm_avail_table_valid_p (dbf, avblk->av_table, avblk->count)))
+    {
+      GDBM_SET_ERRNO (dbf, GDBM_BAD_AVAIL, TRUE);
+      return -1;
+    }
+  return 0;
+}
+
+int
+gdbm_bucket_avail_table_validate (GDBM_FILE dbf, hash_bucket *bucket)
+{
+  if (!(bucket->av_count >= 0
+	&& bucket->av_count <= BUCKET_AVAIL
+	&& gdbm_avail_table_valid_p (dbf, bucket->bucket_avail,
+				     bucket->av_count)))
+    {
+      GDBM_SET_ERRNO (dbf, GDBM_BAD_AVAIL, TRUE);
+      return -1;
+    }
+  return 0;
+}
+
+static int
+validate_header (gdbm_file_header const *hdr, struct stat const *st)
+{
+  int dir_size, dir_bits;
+  
+  /* Is the magic number good? */
+  if (hdr->header_magic != GDBM_MAGIC)
+    {
+      switch (hdr->header_magic)
+	{
+	case GDBM_OMAGIC:
+	  /* OK */
+	  break;
+
+	case GDBM_OMAGIC_SWAP:
+	case GDBM_MAGIC32_SWAP:
+	case GDBM_MAGIC64_SWAP:
+	  return GDBM_BYTE_SWAPPED;
+
+	case GDBM_MAGIC32:
+	case GDBM_MAGIC64:
+	  return GDBM_BAD_FILE_OFFSET;
+
+	default:
+	  return GDBM_BAD_MAGIC_NUMBER;
+	}
+    }
+  
+  if (!(hdr->block_size > 0
+	&& hdr->block_size > sizeof (gdbm_file_header)
+	&& hdr->block_size - sizeof (gdbm_file_header) >=
+	sizeof(hdr->avail.av_table[0])))
+    {
+      return GDBM_BLOCK_SIZE_ERROR;
+    }
+
+  if (hdr->next_block != st->st_size)
+    /* FIXME: Should return GDBM_NEED_RECOVERY instead? */
+    return GDBM_BAD_HEADER;
+
+  /* Make sure dir and dir + dir_size fall within the file boundary */
+  if (!(hdr->dir > 0
+	&& hdr->dir < st->st_size
+	&& hdr->dir_size > 0
+	&& hdr->dir + hdr->dir_size < st->st_size))
+    return GDBM_BAD_HEADER;
+
+  compute_directory_size (hdr->block_size, &dir_size, &dir_bits);
+  if (!(hdr->dir_size >= dir_size))
+    return GDBM_BAD_HEADER;
+  compute_directory_size (hdr->dir_size, &dir_size, &dir_bits);
+  if (hdr->dir_bits != dir_bits)
+    return GDBM_BAD_HEADER;
+  
+  if (!(hdr->bucket_size > sizeof(hash_bucket)))
+    return GDBM_BAD_HEADER;
+
+  if (hdr->bucket_elems != bucket_element_count (hdr->bucket_size))
+    return GDBM_BAD_HEADER;
+
+  if (((hdr->block_size - sizeof (gdbm_file_header)) / sizeof(avail_elem) + 1)
+      != hdr->avail.size)
+    return GDBM_BAD_HEADER;    
+  
+  return 0;
+}
+  
 GDBM_FILE 
 gdbm_fd_open (int fd, const char *file_name, int block_size,
 	      int flags, void (*fatal_func) (const char *))
@@ -194,7 +312,7 @@ gdbm_fd_open (int fd, const char *file_name, int block_size,
 	  block_size = STATBLKSIZE (file_stat);
 	  flags &= ~GDBM_BSEXACT;
 	}
-      compute_directory_size (dbf, block_size, &dir_size, &dir_bits);
+      compute_directory_size (block_size, &dir_size, &dir_bits);
       GDBM_DEBUG (GDBM_DEBUG_OPEN, "%s: computed dir_size=%d, dir_bits=%d",
 		  dbf->name, dir_size, dir_bits);
       /* Check for correct block_size. */
@@ -245,9 +363,7 @@ gdbm_fd_open (int fd, const char *file_name, int block_size,
       dbf->header->dir = dbf->header->block_size;
 
       /* Create the first and only hash bucket. */
-      dbf->header->bucket_elems =
-	(dbf->header->block_size - sizeof (hash_bucket))
-	/ sizeof (bucket_element) + 1;
+      dbf->header->bucket_elems = bucket_element_count (dbf->header->block_size);
       dbf->header->bucket_size  = dbf->header->block_size;
       dbf->bucket = calloc (1, dbf->header->bucket_size);
       if (dbf->bucket == NULL)
@@ -311,9 +427,20 @@ gdbm_fd_open (int fd, const char *file_name, int block_size,
 	  SAVE_ERRNO (gdbm_close (dbf));
 	  return NULL;
 	}
-
+      
+      if (_gdbm_file_extend (dbf, dbf->header->next_block))
+	{
+	  GDBM_DEBUG (GDBM_DEBUG_OPEN|GDBM_DEBUG_ERR,
+		      "%s: error extending file: %s",
+		      dbf->name, gdbm_db_strerror (dbf));	  
+	  if (!(flags & GDBM_CLOERROR))
+	    dbf->desc = -1;
+	  SAVE_ERRNO (gdbm_close (dbf));
+	  return NULL;
+	}
+	  
       /* Wait for initial configuration to be written to disk. */
-      __fsync (dbf);
+      gdbm_file_sync (dbf);
 
       free (dbf->bucket);
     }
@@ -323,7 +450,8 @@ gdbm_fd_open (int fd, const char *file_name, int block_size,
 	 header and initialize the hash directory. */
 
       gdbm_file_header partial_header;  /* For the first part of it. */
-
+      int rc;
+      
       /* Read the partial file header. */
       if (_gdbm_full_read (dbf, &partial_header, sizeof (gdbm_file_header)))
 	{
@@ -336,33 +464,17 @@ gdbm_fd_open (int fd, const char *file_name, int block_size,
 	  return NULL;
 	}
 
-      /* Is the magic number good? */
-      if (partial_header.header_magic != GDBM_MAGIC
-	  && partial_header.header_magic != GDBM_OMAGIC)
+      /* Is the header valid? */
+      rc = validate_header (&partial_header, &file_stat);
+      if (rc != GDBM_NO_ERROR)
 	{
 	  if (!(flags & GDBM_CLOERROR))
 	    dbf->desc = -1;
 	  gdbm_close (dbf);
-	  switch (partial_header.header_magic)
-	    {
-	      case GDBM_OMAGIC_SWAP:
-	      case GDBM_MAGIC32_SWAP:
-	      case GDBM_MAGIC64_SWAP:
-		GDBM_SET_ERRNO2 (NULL, GDBM_BYTE_SWAPPED, FALSE,
-				 GDBM_DEBUG_OPEN);
-		break;
-	      case GDBM_MAGIC32:
-	      case GDBM_MAGIC64:
-		GDBM_SET_ERRNO2 (NULL, GDBM_BAD_FILE_OFFSET, FALSE,
-				 GDBM_DEBUG_OPEN);
-		break;
-	      default:
-		GDBM_SET_ERRNO2 (NULL, GDBM_BAD_MAGIC_NUMBER, FALSE,
-				 GDBM_DEBUG_OPEN);
-	    }
+	  GDBM_SET_ERRNO2 (NULL, rc, FALSE, GDBM_DEBUG_OPEN);
 	  return NULL;
 	}
-
+      
       /* It is a good database, read the entire header. */
       dbf->header = malloc (partial_header.block_size);
       if (dbf->header == NULL)
@@ -373,6 +485,7 @@ gdbm_fd_open (int fd, const char *file_name, int block_size,
 	  GDBM_SET_ERRNO2 (NULL, GDBM_MALLOC_ERROR, FALSE, GDBM_DEBUG_OPEN);
 	  return NULL;
 	}
+      
       memcpy (dbf->header, &partial_header, sizeof (gdbm_file_header));
       if (_gdbm_full_read (dbf, &dbf->header->avail.av_table[1],
 			   dbf->header->block_size - sizeof (gdbm_file_header)))
@@ -385,6 +498,14 @@ gdbm_fd_open (int fd, const char *file_name, int block_size,
 	  SAVE_ERRNO (gdbm_close (dbf));
 	  return NULL;
 	}
+
+      if (gdbm_avail_block_validate (dbf, &dbf->header->avail))
+	{
+	  if (!(flags & GDBM_CLOERROR))
+	    dbf->desc = -1;
+	  SAVE_ERRNO (gdbm_close (dbf));
+	  return NULL;
+	}	  
 	
       /* Allocate space for the hash table directory.  */
       dbf->dir = malloc (dbf->header->dir_size);
@@ -398,7 +519,7 @@ gdbm_fd_open (int fd, const char *file_name, int block_size,
 	}
 
       /* Read the hash table directory. */
-      file_pos = __lseek (dbf, dbf->header->dir, SEEK_SET);
+      file_pos = gdbm_file_seek (dbf, dbf->header->dir, SEEK_SET);
       if (file_pos != dbf->header->dir)
 	{
 	  if (!(flags & GDBM_CLOERROR))
@@ -524,8 +645,7 @@ _gdbm_init_cache (GDBM_FILE dbf, size_t size)
 
   if (dbf->bucket_cache == NULL)
     {
-      dbf->bucket_cache = GDBM_DEBUG_ALLOC ("_gdbm_init_cache:malloc-failure",
-	                      malloc (sizeof(cache_elem) * size));
+      dbf->bucket_cache = calloc (size, sizeof(cache_elem));
       if (dbf->bucket_cache == NULL)
         {
           GDBM_SET_ERRNO (dbf, GDBM_MALLOC_ERROR, TRUE);
@@ -535,22 +655,28 @@ _gdbm_init_cache (GDBM_FILE dbf, size_t size)
 
       for (index = 0; index < size; index++)
         {
-	  (dbf->bucket_cache[index]).ca_bucket = 
-	    GDBM_DEBUG_ALLOC ("_gdbm_init_cache:bucket-malloc-failure",
-	                      malloc (dbf->header->bucket_size));
+	  (dbf->bucket_cache[index]).ca_bucket =
+	    malloc (dbf->header->bucket_size);
           if ((dbf->bucket_cache[index]).ca_bucket == NULL)
 	    {
               GDBM_SET_ERRNO (dbf, GDBM_MALLOC_ERROR, TRUE);
 	      return -1;
             }
-          (dbf->bucket_cache[index]).ca_adr = 0;
-          (dbf->bucket_cache[index]).ca_changed = FALSE;
-          (dbf->bucket_cache[index]).ca_data.hash_val = -1;
-          (dbf->bucket_cache[index]).ca_data.elem_loc = -1;
-          (dbf->bucket_cache[index]).ca_data.dptr = NULL;
+	  dbf->bucket_cache[index].ca_data.dptr = NULL;
+	  dbf->bucket_cache[index].ca_data.dsize = 0;
+	  _gdbm_cache_entry_invalidate (dbf, index);
         }
       dbf->bucket = dbf->bucket_cache[0].ca_bucket;
       dbf->cache_entry = &dbf->bucket_cache[0];
     }
   return 0;
 }
+
+void
+_gdbm_cache_entry_invalidate (GDBM_FILE dbf, int index)
+{
+  dbf->bucket_cache[index].ca_adr = 0;
+  dbf->bucket_cache[index].ca_changed = FALSE;
+  dbf->bucket_cache[index].ca_data.hash_val = -1;
+  dbf->bucket_cache[index].ca_data.elem_loc = -1;
+}
