@@ -44,8 +44,15 @@
 # define _GDBM_NEED_REMAP(dbf) \
   (!(dbf)->mapped_region || (dbf)->mapped_pos == (dbf)->mapped_size)
 /* Return the sum of the currently mapped size and DELTA */
-# define SUM_FILE_SIZE(dbf, delta) \
-  ((dbf)->mapped_off + (dbf)->mapped_size + (delta))
+static inline off_t
+SUM_FILE_SIZE (GDBM_FILE dbf, off_t delta)
+{
+  if (delta >= 0
+      && off_t_sum_ok (dbf->mapped_off, dbf->mapped_size)
+      && off_t_sum_ok (dbf->mapped_off + dbf->mapped_size, delta))
+    return dbf->mapped_off + dbf->mapped_size + delta;
+  return -1;
+}
 
 /* Store the size of the GDBM file DBF in *PSIZE.
    Return 0 on success and -1 on failure. */
@@ -115,48 +122,6 @@ _gdbm_internal_remap (GDBM_FILE dbf, size_t size)
   return 0;
 }
 
-/* Grow the disk file of DBF to the SIZE bytes in length. Fill the
-   newly allocated space with zeros. */
-static int
-_gdbm_file_extend (GDBM_FILE dbf, off_t size)
-{
-  size_t page_size = sysconf (_SC_PAGESIZE);
-  char *buf;
-  off_t file_end;
-
-  file_end = lseek (dbf->desc, 0, SEEK_END);
-  if (!file_end)
-    {
-      GDBM_SET_ERRNO (dbf, GDBM_FILE_SEEK_ERROR, FALSE);
-      return -1;
-    }
-  size -= file_end;
-  
-  if (size < page_size)
-    page_size = size;
-  buf = calloc (1, page_size);
-  if (!buf)
-    {
-      GDBM_SET_ERRNO (dbf, GDBM_MALLOC_ERROR, FALSE);
-      return -1;
-    }
-
-  while (size)
-    {
-      ssize_t n = write (dbf->desc, buf, size < page_size ? size : page_size);
-      if (n <= 0)
-	{
-	  GDBM_SET_ERRNO (dbf, GDBM_FILE_WRITE_ERROR, FALSE);
-	  break;
-	}
-      size -= n;
-    }
-  free (buf);
-  if (size)
-    return -1;
-  return 0;
-}
-  
 # define _REMAP_DEFAULT 0
 # define _REMAP_EXTEND  1
 # define _REMAP_END     2
@@ -182,6 +147,17 @@ _gdbm_mapped_remap (GDBM_FILE dbf, off_t size, int flag)
 {
   off_t file_size, pos;
 
+  if (size < 0)
+    {
+      errno = EINVAL;
+      GDBM_SET_ERRNO (dbf, GDBM_FILE_SEEK_ERROR, TRUE);
+      return -1;
+    }
+
+  if (size < dbf->mapped_size)
+    /* Nothing to do */
+    return 0;
+  
   if (_gdbm_file_size (dbf, &file_size))
     {
       SAVE_ERRNO (_gdbm_mapped_unmap (dbf));
@@ -205,7 +181,6 @@ _gdbm_mapped_remap (GDBM_FILE dbf, off_t size, int flag)
 	    }
 	  else
 	    {
-	      size = file_size;
 	      return 0;
 	    }
 	}
@@ -233,7 +208,12 @@ _gdbm_mapped_remap (GDBM_FILE dbf, off_t size, int flag)
       dbf->mapped_pos += dbf->mapped_off;
       dbf->mapped_off = 0;
     }
-
+  if (pos > file_size)
+    {
+      errno = EINVAL;
+      GDBM_SET_ERRNO (dbf, GDBM_FILE_SEEK_ERROR, TRUE);
+      return -1;
+    }
   return _gdbm_internal_remap (dbf, size);
 }
 
@@ -271,6 +251,10 @@ _gdbm_mapped_read (GDBM_FILE dbf, void *buffer, size_t len)
 		{
 		  int rc;
 
+		  if (dbf->need_recovery)
+		    return -1;
+
+		  /* Disable memory mapping and retry */
 		  dbf->memory_mapping = FALSE;
 		  if (lseek (dbf->desc, pos, SEEK_SET) != pos)
 		    return total > 0 ? total : -1;
@@ -321,6 +305,9 @@ _gdbm_mapped_write (GDBM_FILE dbf, void *buffer, size_t len)
 		{
 		  int rc;
 
+		  if (dbf->need_recovery)
+		    return -1;
+
 		  dbf->memory_mapping = FALSE;
 		  if (lseek (dbf->desc, pos, SEEK_SET) != pos)
 		    return total > 0 ? total : -1;
@@ -382,6 +369,12 @@ _gdbm_mapped_lseek (GDBM_FILE dbf, off_t offset, int whence)
 	  }
 	}
 
+      if (needle < 0)
+	{
+	  errno = EINVAL;
+	  return -1;
+	}
+      
       if (!_GDBM_IN_MAPPED_REGION_P (dbf, needle))
 	{
 	  _gdbm_mapped_unmap (dbf);
