@@ -29,7 +29,7 @@ int _gdbm_write_bucket (GDBM_FILE, cache_elem *);
 
 /* From falloc.c */
 off_t _gdbm_alloc       (GDBM_FILE, int);
-void _gdbm_free         (GDBM_FILE, off_t, int);
+int  _gdbm_free         (GDBM_FILE, off_t, int);
 int  _gdbm_put_av_elem  (avail_elem, avail_elem [], int *, int);
 
 /* From findkey.c */
@@ -48,6 +48,10 @@ void _gdbm_fatal	(GDBM_FILE, const char *);
 
 /* From gdbmopen.c */
 int _gdbm_init_cache	(GDBM_FILE, size_t);
+void _gdbm_cache_entry_invalidate (GDBM_FILE, int);
+
+int gdbm_avail_block_validate (GDBM_FILE dbf, avail_block *avblk);
+int gdbm_bucket_avail_table_validate (GDBM_FILE dbf, hash_bucket *bucket);
 
 /* From mmap.c */
 int _gdbm_mapped_init	(GDBM_FILE);
@@ -64,6 +68,7 @@ int _gdbm_lock_file	(GDBM_FILE);
 /* From fullio.c */
 int _gdbm_full_read (GDBM_FILE, void *, size_t);
 int _gdbm_full_write (GDBM_FILE, void *, size_t);
+int _gdbm_file_extend (GDBM_FILE dbf, off_t size);
 
 /* From base64.c */
 int _gdbm_base64_encode (const unsigned char *input, size_t input_len,
@@ -79,4 +84,47 @@ int _gdbm_dump (GDBM_FILE dbf, FILE *fp);
 /* From recover.c */
 int _gdbm_next_bucket_dir (GDBM_FILE dbf, int bucket_dir);
 
+/* I/O functions */
+static inline ssize_t
+gdbm_file_read (GDBM_FILE dbf, void *buf, size_t size)
+{
+#if HAVE_MMAP
+  return _gdbm_mapped_read (dbf, buf, size);
+#else
+  return read (dbf->desc, buf, size);
+#endif
+}
+
+static inline ssize_t
+gdbm_file_write (GDBM_FILE dbf, void *buf, size_t size)
+{
+#if HAVE_MMAP
+  return _gdbm_mapped_write (dbf, buf, size);
+#else
+  return write (dbf->desc, buf, size);
+#endif
+}
+
+static inline off_t
+gdbm_file_seek (GDBM_FILE dbf, off_t off, int whence)
+{
+#if HAVE_MMAP
+  return _gdbm_mapped_lseek (dbf, off, whence);
+#else
+  return lseek (dbf->desc, off, whence);
+#endif
+}
+
+static inline void
+gdbm_file_sync (GDBM_FILE dbf)
+{
+#if HAVE_MMAP
+  _gdbm_mapped_sync (dbf);
+#elif HAVE_FSYNC
+  fsync (dbf->desc);
+#else
+  sync ();
+  sync ();
+#endif
+}
 
